"""
Module provides helper functions and data management classes for visualizing
scenarios.
"""
import inspect
from dataclasses import dataclass
from typing import Optional, Union, List, Tuple

import matplotlib as mpl
import matplotlib.pyplot as plt
import networkx.drawing.nx_pylab as nxp
import numpy as np
from scipy.interpolate import CubicSpline

from ..serialization import COLOR_SCHEMES_ID, JsonSerializable, serializable
from ..simulation.scada.scada_data import ScadaData

# Selection of functions for processing visualization data
stat_funcs = {
    'mean': np.mean,
    'min': np.min,
    'max': np.max
}


@dataclass
class JunctionObject:
    """
    Represents a junction component (e.g. nodes, tanks, reservoirs, ...) in a
    water distribution network and manages all relevant attributes for drawing.

    Attributes
    ----------
    nodelist : `list`
        List of all nodes in WDN pertaining to this component type.
    pos : `dict`
        A dictionary mapping nodes to their coordinates in the correct format
        for drawing.
    node_shape : :class:`matplotlib.path.Path` or None
        A shape representing the object, if none, the networkx default circle
        is used.
    node_size : `int`, default = 10
        The size of each node.
    node_color : `str` or `list`, default = 'k'
        If `string`: the color for all nodes, if `list`: a list of lists
        containing a numerical value for each node per frame, which will be
        used for coloring.
    interpolated : `bool`, default = False
        Set to True, if node_colors are interpolated for smoother animation.
    """
    nodelist: list
    pos: dict
    node_shape: mpl.path.Path = None
    node_size: int = 10
    node_color: Union[str, list] = 'k'
    interpolated: bool = False

    def add_frame(self, statistic: str, values: np.ndarray,
                  pit: int, intervals: Union[int, List[Union[int, float]]]):
        """
        Adds a new frame of node_color based on a given statistic.

        Parameters
        ----------
        statistic : `str`
            The statistic to calculate for the data. Can be 'mean', 'min',
             'max' or 'time_step'.
        values : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
            The node values over time as extracted from the scada data.
        pit : `int`
            The point in time for the 'time_step' statistic.
        intervals : `int`, `list[int]` or `list[float]`
            If provided, the data will be grouped into intervals. It can be an
            integer specifying the number of groups or a list of boundary
            points.

        Raises
        ------
        ValueError
            If interval, pit or statistic is not correctly provided.

        """
        if statistic in stat_funcs:
            stat_values = stat_funcs[statistic](values, axis=0)
        elif statistic == 'time_step':
            if not pit and pit != 0:
                raise ValueError(
                    'Please input point in time (pit) parameter when selecting'
                    ' time_step statistic')
            stat_values = np.take(values, pit, axis=0)
        else:
            raise ValueError(
                'Statistic parameter must be mean, min, max or time_step')

        if intervals is None:
            pass
        elif isinstance(intervals, (int, float)):
            interv = np.linspace(stat_values.min(), stat_values.max(),
                                 intervals + 1)
            stat_values = np.digitize(stat_values, interv) - 1
        elif isinstance(intervals, list):
            stat_values = np.digitize(stat_values, intervals) - 1
        else:
            raise ValueError(
                'Intervals must be either number of groups or list of interval'
                ' boundary points')

        sorted_values = [v for _, v in zip(self.nodelist, stat_values)]

        if isinstance(self.node_color, str):
            # First run of this method
            self.node_color = []
            self.vmin = min(sorted_values)
            self.vmax = max(sorted_values)

        self.node_color.append(sorted_values)
        self.vmin = min(*sorted_values, self.vmin)
        self.vmax = max(*sorted_values, self.vmax)

    def get_frame(self, frame_number: int = 0):
        """
        Returns all attributes necessary for networkx to draw the specified
        frame.

        Parameters
        ----------
        frame_number : `int`, default = 0
            The frame whose parameters should be returned. Default is 0, this
            is also used if only 1 frame exists (e.g. for plots, not
            animations).

        Returns
        -------
        valid_params : `dict`
            A dictionary containing all attributes that function as parameters
            for `networkx.drawing.nx_pylab.draw_networkx_nodes() <https://networkx.org/documentation/stable/reference/generated/networkx.drawing.nx_pylab.draw_networkx_nodes.html#draw-networkx-nodes>`_.
        """

        attributes = vars(self).copy()

        if not isinstance(self.node_color, str):
            if self.interpolated:
                if frame_number > len(self.node_color_inter):
                    frame_number = -1
                attributes['node_color'] = self.node_color_inter[frame_number]
            else:
                if frame_number > len(self.node_color):
                    frame_number = -1
                attributes['node_color'] = self.node_color[frame_number]

        sig = inspect.signature(nxp.draw_networkx_nodes)

        valid_params = {
            key: value for key, value in attributes.items()
            if key in sig.parameters and value is not None
        }

        return valid_params

    def get_frame_mask(self, mask, color):
        """
        Returns all attributes necessary for networkx to draw the specified
        frame mask. Meaning covering all masked junction objects with the
        default value.

        Parameters
        ----------
        mask: `np.ndarray`
            An array consisting of 0 and 1, where 0 means no sensor. Nodes
            without sensor are to be masked.
        color:
            The default color of masked nodes.

        Returns
        -------
        valid_params : `dict`
            A dictionary containing all attributes that function as parameters
            for `networkx.drawing.nx_pylab.draw_networkx_nodes() <https://networkx.org/documentation/stable/reference/generated/networkx.drawing.nx_pylab.draw_networkx_nodes.html#draw-networkx-nodes>`_.
        """

        attributes = vars(self).copy()

        attributes['nodelist'] = [node for node, flag in
                                  zip(self.nodelist, mask) if not flag]
        attributes['node_color'] = color

        sig = inspect.signature(nxp.draw_networkx_nodes)

        valid_params = {
            key: value for key, value in attributes.items()
            if key in sig.parameters and key not in ['vmin', 'vmax', 'cmap']
            and value is not None
        }

        return valid_params

    def interpolate(self, num_inter_frames: int):
        """
        Interpolates node_color values for smoother animations.

        Parameters
        ----------
        num_inter_frames : `int`
            The number of total frames after interpolation.
        """
        if isinstance(self.node_color, str) or len(self.node_color) <= 1:
            return

        tmp_node_color = np.array(self.node_color)
        steps, num_nodes = tmp_node_color.shape

        x_axis = np.linspace(0, steps - 1, steps)
        new_x_axis = np.linspace(0, steps - 1, num_inter_frames)

        self.node_color_inter = np.zeros(((len(new_x_axis)), num_nodes))

        for node in range(num_nodes):
            cs = CubicSpline(x_axis, tmp_node_color[:, node])
            self.node_color_inter[:, node] = cs(new_x_axis)

        self.interpolated = True

    def add_attributes(self, attributes: dict):
        """
        Adds the given attributes dict as class attributes.

        Parameters
        ----------
        attributes : `dict`
            Attributes dict, which is to be added as class attributes.
        """
        for key, value in attributes.items():
            setattr(self, key, value)


@dataclass
class EdgeObject:
    """
    Represents an edge component (pipes) in a water distribution network and
    manages all relevant attributes for drawing.

    Attributes
    ----------
    edgelist : `list`
        List of all edges in WDN pertaining to this component type.
    pos : `dict`
        A dictionary mapping pipes to their coordinates in the correct format
        for drawing.
    edge_color : `str` or `list`, default = 'k'
        If `string`: the color for all edges, if `list`: a list of lists
        containing a numerical value for each edge per frame, which will be
        used for coloring.
    interpolated : `dict`, default = {}
        Filled with interpolated frames if interpolation method is called.
    """
    edgelist: list
    pos: dict
    edge_color: Union[str, list] = 'k'
    interpolated = {}

    def rescale_widths(self, line_widths: Tuple[int, int] = (1, 2)):
        """
        Rescales all edge widths to the given interval.

        Parameters
        ----------
        line_widths : `Tuple[int]`, default = (1, 2)
        Min and max value, to which the edge widths should be scaled.

        Raises
        ------
        AttributeError
            If no edge width attribute exists yet.
        """
        if not hasattr(self, 'width'):
            raise AttributeError(
                'Please call add_frame with edge_param=width before rescaling'
                ' the widths.')

        vmin = min(min(l) for l in self.width)
        vmax = max(max(l) for l in self.width)

        tmp = []
        for il in self.width:
            tmp.append(
                self.__rescale(il, line_widths, values_min_max=(vmin, vmax)))
        self.width = tmp

    def add_frame(
            self, topology, edge_param: str,
            scada_data: Optional[ScadaData],
            parameter: str = 'flow_rate', statistic: str = 'mean',
            pit: Optional[Union[int, Tuple[int]]] = None,
            species: str = None,
            intervals: Optional[Union[int, List[Union[int, float]]]] = None,
            use_sensor_data: bool = None):
        """
        Adds a new frame of edge_color or edge width based on the given data
        and statistic.

        Parameters
        ----------
        topology : :class:`~epyt_flow.topology.NetworkTopology`
            Topology object retrieved from the scenario, containing the
            structure of the water distribution network.
        edge_param : `str`
            Method can be called with edge_width or edge_color to calculate
            either the width or color for the next frame.
        scada_data : :class:`~epyt_flow.simulation.scada.scada_data.ScadaData`
            SCADA data created by the :class:`~epyt_flow.simulation.scenario_simulator.ScenarioSimulator`
            instance, is used to retrieve data for the next frame.
        parameter : `str`, default = 'flow_rate'
            The link data to visualize. Options are 'flow_rate', 'velocity', or
            'status'. Default is 'flow_rate'.
        statistic : `str`, default = 'mean'
            The statistic to calculate for the data. Can be 'mean', 'min',
             'max' or 'time_step'.
        pit : `int`
            The point in time for the 'time_step' statistic.
        species: `str`, optional
            Key of the species. Necessary only for parameter
            'bulk_species_concentration'.
        intervals : `int`, `list[int]` or `list[float]`
            If provided, the data will be grouped into intervals. It can be an
            integer specifying the number of groups or a list of boundary
            points.
        use_sensor_data : `bool`, optional
            If `True`, instead of using raw simulation data, the data recorded
            by the corresponding sensors in the system is used for the
            visualization. Note: Not all components may have a sensor attached
            and sensors may be subject to sensor faults or noise.

        Raises
        ------
        ValueError
            If parameter, interval, pit or statistic is not set correctly.
        """
        if edge_param == 'edge_width' and not hasattr(self, 'width'):
            self.width = []
        elif edge_param == 'edge_color':
            if isinstance(self.edge_color, str):
                self.edge_color = []
                self.edge_vmin = float('inf')
                self.edge_vmax = float('-inf')

        if parameter == 'flow_rate':
            if use_sensor_data:
                values, self.mask = scada_data.get_data_flows_as_edge_features()
                values = values[:, ::2]
                self.mask = self.mask[::2]
            else:
                values = scada_data.flow_data_raw
        elif parameter == 'link_quality':
            if use_sensor_data:
                values, self.mask = scada_data.get_data_links_quality_as_edge_features()
                values = values[:, ::2]
                self.mask = self.mask[::2]
            else:
                values = scada_data.link_quality_data_raw
        elif parameter == 'custom_data':
            values = scada_data
        elif parameter == 'bulk_species_concentration':
            if not species:
                raise ValueError('Species must be given when using '
                                 'bulk_species_concentration')
            if use_sensor_data:
                values, self.mask = scada_data.get_data_bulk_species_concentrations_as_edge_features()
                self.mask = self.mask[::2,
                            scada_data.sensor_config.bulk_species.index(
                                species)]
                values = values[:, ::2,
                         scada_data.sensor_config.bulk_species.index(species)]
            else:
                values = scada_data.bulk_species_link_concentration_raw[:,
                         scada_data.sensor_config.bulk_species.index(species),
                         :]
        elif parameter == 'diameter':
            value_dict = {
                link[0]: topology.get_link_info(link[0])['diameter'] for
                link in topology.get_all_links()}
            sorted_values = [value_dict[x[0]] for x in
                             topology.get_all_links()]

            if edge_param == 'edge_width':
                self.width.append(sorted_values)
            else:
                self.edge_color.append(sorted_values)
                self.edge_vmin = min(*sorted_values, self.edge_vmin)
                self.edge_vmax = max(*sorted_values, self.edge_vmax)
            return
        else:
            raise ValueError('Parameter must be flow_rate, link_quality, '
                             'diameter or custom_data.')

        if statistic in stat_funcs:
            stat_values = stat_funcs[statistic](values, axis=0)
        elif statistic == 'time_step':
            if not pit and pit != 0:
                raise ValueError(
                    'Please input point in time (pit) parameter when selecting'
                    ' time_step statistic')
            stat_values = np.take(values, pit, axis=0)
        else:
            raise ValueError(
                'Statistic parameter must be mean, min, max or time_step')

        if intervals is None:
            pass
        elif isinstance(intervals, (int, float)):
            interv = np.linspace(stat_values.min(), stat_values.max(),
                                 intervals + 1)
            stat_values = np.digitize(stat_values, interv) - 1
        elif isinstance(intervals, list):
            stat_values = np.digitize(stat_values, intervals) - 1
        else:
            raise ValueError(
                'Intervals must be either number of groups or list of interval'
                ' boundary points')

        sorted_values = list(stat_values)

        if edge_param == 'edge_width':
            self.width.append(sorted_values)
        else:
            self.edge_color.append(sorted_values)
            self.edge_vmin = min(*sorted_values, self.edge_vmin)
            self.edge_vmax = max(*sorted_values, self.edge_vmax)

    def get_frame(self, frame_number: int = 0):
        """
        Returns all attributes necessary for networkx to draw the specified
        frame.

        Parameters
        ----------
        frame_number : `int`, default = 0
            The frame whose parameters should be returned. Default is 0, this
            is also used if only 1 frame exists (e.g. for plots, not
            animations).

        Returns
        -------
        valid_params : `dict`
            A dictionary containing all attributes that function as parameters
            for `networkx.drawing.nx_pylab.draw_networkx_edges() <https://networkx.org/documentation/stable/reference/generated/networkx.drawing.nx_pylab.draw_networkx_edges.html#draw-networkx-edges>`_.
        """
        attributes = vars(self).copy()

        if not isinstance(self.edge_color, str):
            if 'edge_color' in self.interpolated.keys():
                if frame_number > len(self.interpolated['edge_color']):
                    frame_number = -1
                attributes['edge_color'] = self.interpolated['edge_color'][
                    frame_number]
            else:
                if frame_number > len(self.edge_color):
                    frame_number = -1
                attributes['edge_color'] = self.edge_color[frame_number]

        if hasattr(self, 'width'):
            if 'width' in self.interpolated.keys():
                if frame_number > len(self.interpolated['width']):
                    frame_number = -1
                attributes['width'] = self.interpolated['width'][frame_number]
            else:
                if frame_number > len(self.width):
                    frame_number = -1
                attributes['width'] = self.width[frame_number]

        sig = inspect.signature(nxp.draw_networkx_edges)

        valid_params = {
            key: value for key, value in attributes.items()
            if key in sig.parameters and value is not None
        }

        return valid_params

    def get_frame_mask(self, frame_number: int = 0, color='k'):
        """
        Returns all attributes necessary for networkx to draw the specified
        frame mask.

        Parameters
        ----------
        frame_number : `int`, default = 0
            The frame whose parameters should be returned. Default is 0, this
            is also used if only 1 frame exists (e.g. for plots, not
            animations).
        color:
            The default color of masked nodes.

        Returns
        -------
        valid_params : `dict`
            A dictionary containing all attributes that function as parameters
            for `networkx.drawing.nx_pylab.draw_networkx_edges() <https://networkx.org/documentation/stable/reference/generated/networkx.drawing.nx_pylab.draw_networkx_edges.html#draw-networkx-edges>`_.
        """
        attributes = vars(self).copy()

        attributes['edgelist'] = [edge for edge, flag in
                                  zip(self.edgelist, self.mask) if not flag]
        attributes['edge_color'] = color

        if hasattr(self, 'width'):
            if 'width' in self.interpolated.keys():
                if frame_number > len(self.interpolated['width']):
                    frame_number = -1
                attributes['width'] = self.interpolated['width'][frame_number]
            else:
                if frame_number > len(self.width):
                    frame_number = -1
                attributes['width'] = self.width[frame_number]
            attributes['width'] = [edge for edge, flag in
                                   zip(attributes['width'].copy(), self.mask)
                                   if not flag]

        sig = inspect.signature(nxp.draw_networkx_edges)

        valid_params = {
            key: value for key, value in attributes.items()
            if key in sig.parameters and key not in ['vmin', 'vmax', 'cmap']
            and value is not None
        }

        return valid_params

    def interpolate(self, num_inter_frames: int):
        """
        Interpolates edge_color and width values for smoother animations.

        Parameters
        ----------
        num_inter_frames : `int`
            The number of total frames after interpolation.
        """
        targets = {'edge_color': self.edge_color}
        if hasattr(self, 'width'):
            targets['width'] = self.width

        for name, inter_target in targets.items():
            if isinstance(inter_target, str) or len(inter_target) <= 1:
                continue

            tmp_target = np.array(inter_target)
            steps, num_edges = tmp_target.shape

            x_axis = np.linspace(0, steps - 1, steps)
            new_x_axis = np.linspace(0, steps - 1, num_inter_frames)

            vals_inter = np.zeros(((len(new_x_axis)), num_edges))

            for edge in range(num_edges):
                cs = CubicSpline(x_axis, tmp_target[:, edge])
                vals_inter[:, edge] = cs(new_x_axis)

            self.interpolated[name] = vals_inter

    def add_attributes(self, attributes: dict):
        """
        Adds the given attributes dict as class attributes.

        Parameters
        ----------
        attributes : `dict`
            Attributes dict, which is to be added as class attributes.
        """
        for key, value in attributes.items():
            setattr(self, key, value)

    def __rescale(self, values: Union[np.ndarray, list],
                  scale_min_max: Union[List, Tuple[int]],
                  values_min_max: Union[
                      List, Tuple[int, int]] = None) -> np.ndarray:
        """
        Rescales the given values to a new range.

        This method rescales an array of values to fit within a specified
        minimum and maximum scale range. Optionally, the minimum and maximum
        of the input values can be manually provided; otherwise, they are
        automatically determined from the data.

        Parameters
        ----------
        values : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_ or `list`
            The array of numerical values to be rescaled.
        scale_min_max : `list` or `tuple`
            A list or tuple containing two elements: the minimum and maximum
            values of the desired output range.
        values_min_max : `list` or `tuple`, optional
            A list or tuple containing two elements: the minimum and maximum
            values of the input data. If not provided, they are computed from
            the input `values`. Default is `None`.

        Returns
        -------
        rescaled_values : `list`
            A list of values rescaled to the range specified by
            `scale_min_max`.
        """
        if not values_min_max:
            min_val, max_val = min(values), max(values)
        else:
            min_val, max_val = values_min_max
        scale = scale_min_max[1] - scale_min_max[0]

        def range_map(x):
            return scale_min_max[0] + (x - min_val) / (
                    max_val - min_val) * scale

        vectorized_range_map = np.vectorize(range_map)
        rescaled_widths = vectorized_range_map(np.array(values))

        if hasattr(self, 'mask'):
            rescaled_widths = np.where(self.mask == 1, rescaled_widths, 1.0)

        return rescaled_widths


@serializable(COLOR_SCHEMES_ID, ".epyt_flow_color_scheme")
class ColorScheme(JsonSerializable):
    """
    A class containing the color scheme for the
    :class:`~epyt_flow.visualization.ScenarioVisualizer`.
    """

    def __init__(self, pipe_color: str, node_color: str, pump_color: str,
                 tank_color: str, reservoir_color: str,
                 valve_color: str) -> None:
        """Initializes the ColorScheme class with the given component colors.

        Accepted formats are the string representations accepted by matplotlib:
        https://matplotlib.org/stable/users/explain/colors/colors.html#color-formats

        Parameters
        ----------
        pipe_color : str
            String color format accepted by matplotlib.
        node_color : str
            String color format accepted by matplotlib.
        pump_color : str
            String color format accepted by matplotlib.
        tank_color : str
            String color format accepted by matplotlib.
        reservoir_color : str
            String color format accepted by matplotlib.
        valve_color : str
            String color format accepted by matplotlib.
        """
        self.pipe_color = pipe_color
        self.node_color = node_color
        self.pump_color = pump_color
        self.tank_color = tank_color
        self.reservoir_color = reservoir_color
        self.valve_color = valve_color
        super().__init__()

    def get_attributes(self) -> dict:
        """
        Gets all attributes needed for serialization.

        Returns
        -------
        attr : A dictionary containing all attributes to be serialized.
        """
        attr = {
            k: v for k, v in self.__dict__.items()
            if
            not (k.startswith("__") or k.startswith("_")) and not callable(v)
        }
        return super().get_attributes() | attr

    def __eq__(self, other: any) -> bool:
        """
        Checks if two ColorScheme instances are equal.

        Parameters
        ----------
        other : :class:`~epyt_flow.visualization_utils.ColorScheme`
            The other ColorScheme instance to compare this one with.

        Returns
        -------
        bool
            True if all attributes are the same, False otherwise.
        """
        if not isinstance(other, ColorScheme):
            return False
        return (
                self.pipe_color == other.pipe_color and
                self.node_color == other.node_color and
                self.pump_color == other.pump_color and
                self.tank_color == other.tank_color and
                self.reservoir_color == other.reservoir_color and
                self.valve_color == other.valve_color
        )

    def __str__(self) -> str:
        """
        Returns a string representation of the ColorScheme instance.

        Returns
        -------
        str
            A string describing the ColorScheme instance.
        """
        return (f"ColorScheme(pipe_color={self.pipe_color}, "
                f"node_color={self.node_color}, "
                f"pump_color={self.pump_color}, "
                f"tank_color={self.tank_color}, "
                f"reservoir_color={self.reservoir_color}, "
                f"valve_color={self.valve_color})")


epanet_colors = ColorScheme(
    pipe_color="#0403ee",
    node_color="#0403ee",
    pump_color="#fe00ff",
    tank_color="#02fffd",
    reservoir_color="#00ff00",
    valve_color="#000000"
)

epyt_flow_colors = ColorScheme(
    pipe_color="#29222f",
    node_color="#29222f",
    pump_color="#d79233",
    tank_color="#607b80",
    reservoir_color="#33483d",
    valve_color="#a3320b"
)

black_colors = ColorScheme(
    pipe_color="#000000",
    node_color="#000000",
    pump_color="#000000",
    tank_color="#000000",
    reservoir_color="#000000",
    valve_color="#000000"
)



"""
This module provides REST API handlers for exporting a given SCADA data instance.
"""
import os
from abc import abstractmethod
import warnings
import falcon

from .handlers import ScadaDataBaseHandler
from ...utils import get_temp_folder
from ...simulation.scada import ScadaData, ScadaDataNumpyExport, ScadaDataMatlabExport, \
    ScadaDataXlsxExport


class ScadaDataBaseExportHandler(ScadaDataBaseHandler):
    """
    Base handler for exporting a given SCADA data instance.
    """
    def __init__(self, file_ext: str, **kwds):
        self.__file_ext = file_ext

        super().__init__(**kwds)

    def create_temp_file_path(self, data_id: str, file_ext: str) -> None:
        """
        Returns a path to a temporary file for storing the SCADA data instance.

        Parameters
        ----------
        data_id : `str`
            UUID of the SCADA data.
        file_ext : `str`
            File extension.
        """
        return os.path.join(get_temp_folder(), f"{data_id}.{file_ext}")

    def send_temp_file(self, resp: falcon.Response, tmp_file: str,
                       content_type: str = "application/octet-stream") -> None:
        """
        Sends a given file (`tmp_file`) to the the client.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        tmp_file : `str`
            Path to the temporary file to be send.
        """
        resp.status = falcon.HTTP_200
        resp.content_type = content_type
        with open(tmp_file, 'rb') as f:
            resp.text = f.read()

    @abstractmethod
    def export(self, scada_data: ScadaData, tmp_file: str) -> None:
        """
        Exports a given SCADA data instance to a temporary file.

        Parameters
        ----------
        scada_data : :class:`~epyt_flow.simulation.scada.scada_data.ScadaData`
            SCADA data instance to be exported.
        tmp_file : `str`
            Path to temporary file.
        """
        raise NotImplementedError()

    def on_get(self, _, resp: falcon.Response, data_id: str) -> None:
        """
        Gets the given SCADA data instance.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        data_id : `str`
            UUID of the SCADA data.
        """
        try:
            if self.scada_data_mgr.validate_uuid(data_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            my_scada_data = self.scada_data_mgr.get(data_id)

            tmp_file = self.create_temp_file_path(data_id, self.__file_ext)
            self.export(my_scada_data, tmp_file)

            self.send_temp_file(resp, tmp_file)

            os.remove(tmp_file)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR


class ScadaDataExportHandler(ScadaDataBaseExportHandler):
    """
    Class for handling a GET requests for exporting a given SCADA data instance
    to an .epytflow_scada_data file.
    """
    def __init__(self, **kwds):
        super().__init__(file_ext=".epytflow_scada_data", **kwds)

    def export(self, scada_data: ScadaData, tmp_file: str) -> None:
        scada_data.save_to_file(tmp_file)


class ScadaDataXlsxExportHandler(ScadaDataBaseExportHandler):
    """
    Class for handling a GET requests for exporting a given SCADA data instance to a .xlsx file.
    """
    def __init__(self, **kwds):
        super().__init__(file_ext=".xlsx", **kwds)

    def export(self, scada_data: ScadaData, tmp_file: str) -> None:
        ScadaDataXlsxExport(tmp_file).export(scada_data)


class ScadaDataNumpyExportHandler(ScadaDataBaseExportHandler):
    """
    Class for handling a GET requests for exporting a given SCADA data instance to Numpy data file.
    """
    def __init__(self, **kwds):
        super().__init__(file_ext=".npz", **kwds)

    def export(self, scada_data: ScadaData, tmp_file: str) -> None:
        ScadaDataNumpyExport(tmp_file).export(scada_data)


class ScadaDataMatlabExportHandler(ScadaDataBaseExportHandler):
    """
    Class for handling a GET requests for exporting a given SCADA data instance
    to a Matlab data file.
    """
    def __init__(self, **kwds):
        super().__init__(file_ext=".mat", **kwds)

    def export(self, scada_data: ScadaData, tmp_file: str) -> None:
        ScadaDataMatlabExport(tmp_file).export(scada_data)



"""
The module contains classes for representing simple control rules as used in EPANET.
"""
from typing import Union
from epanet_plus import EpanetConstants

from ..events import ActuatorConstants
from ...serialization import JsonSerializable, SIMPLE_CONTROL_ID, serializable


@serializable(SIMPLE_CONTROL_ID, ".epytflow_simple_control")
class SimpleControlModule(JsonSerializable):
    """
    A class for representing a simple EPANET control rule.

    Parameters
    ----------
    link_id : `str`
        Link ID.
    link_status : `int` or `float`
        Status of the link that is set when the condition is fullfilled.

        Instance of `float` if the link constitutes a pump -- in this case,
        the argument corresponds to the pump speed.

        Instance of `int` if the link constitutes a valve -- in this case,
        must be one of the followig costants defined in
        :class:`~epyt_flow.simulation.events.actuator_events.ActuatorConstants`:

            - EN_CLOSED       = 0
            - EN_OPEN         = 1
            - EN_SET_CLOSED   = -1e10
            - EN_SET_OPEN     = 1e10
    cond_type : `int`
        Condition/Rule type.

        Must be one of the following EPANET constants:

            - EN_LOWLEVEL  = 0
            - EN_HILEVEL   = 1
            - EN_TIMER     = 2
            - EN_TIMEOFDAY = 3
    cond_var_value : `str` or `int`
        Condition/Rule variable value.

        Node ID in the cases of EN_LOWLEVEL or EN_HILEVEL.
        Time of the day (in AM/PM format) in the case of EN_TIMEOFDAY.
        Number of hours (as an integer) since simulation start in the case of EN_TIMER.
    cond_comp_value : `float`
        The condition/rule comparison value at which this control rule is triggered.

        Lower or upper value on the pressure (or tank level) in the cases of
        EN_LOWLEVEL and EN_HILEVEL.

        Will be ignored in all other cases -- i.e. should be set to None.
    """
    def __init__(self, link_id: str, link_status: Union[int, float], cond_type: int,
                 cond_var_value: Union[str, int], cond_comp_value: float,
                 **kwds):
        if not isinstance(link_id, str):
            raise TypeError(f"'link_id' must be an instance of 'str' but not of '{type(link_id)}'")
        if not isinstance(link_status, int) and not isinstance(link_status, float):
            raise TypeError("'link_status' must be an instance of 'int' or 'float' but not " +
                            f"of '{type(link_status)}'")
        if link_status not in [ActuatorConstants.EN_OPEN, ActuatorConstants.EN_CLOSED,
                               ActuatorConstants.EN_SET_OPEN, ActuatorConstants.EN_SET_CLOSED]:
            if link_status < 0:
                raise TypeError("'link_status' can not be negative")
        if cond_type not in [EpanetConstants.EN_TIMEOFDAY, EpanetConstants.EN_TIMER,
                             EpanetConstants.EN_LOWLEVEL, EpanetConstants.EN_HILEVEL]:
            raise ValueError(f"Invalid control type '{cond_type}' in 'cond_type'")

        if cond_type == EpanetConstants.EN_TIMEOFDAY:
            if not isinstance(cond_var_value, str):
                raise TypeError("EN_TIMEOFDAY requires that 'cond_var_value' must be an instance " +
                                f"of 'str' but not of '{type(cond_var_value)}'")
            if not cond_var_value.endswith("AM") and not cond_var_value.endswith("PM"):
                raise ValueError(f"Invalid time of day format '{cond_var_value}' in " +
                                 "'cond_var_value'")
        elif cond_type == EpanetConstants.EN_TIMER:
            if not isinstance(cond_var_value, int):
                raise TypeError("EN_TIMER requires that 'cond_var_value' must be an instance " +
                                f"of 'int' but not of '{type(cond_var_value)}'")
            if cond_var_value < 0:
                raise ValueError("'cond_var_value' can not be negative")
        else:
            if not isinstance(cond_var_value, str):
                raise TypeError("'cond_var_value' must be an instance of 'str' but " +
                                f"not of '{type(cond_var_value)}'")
            if not isinstance(cond_comp_value, float):
                raise TypeError("'cond_comp_value' must be an instance of 'float' " +
                                f"but not of '{type(cond_comp_value)}'")
            if cond_comp_value < 0:
                raise ValueError("'cond_comp_value' can not be negative")

        self.__link_id = link_id
        self.__link_status = link_status
        self.__cond_type = cond_type
        self.__cond_var_value = cond_var_value
        self.__cond_comp_value = cond_comp_value

        super().__init__(**kwds)

    @property
    def link_id(self) -> str:
        """
        Returns the link ID.

        Returns
        -------
        `str`
            Link ID.
        """
        return self.__link_id

    @property
    def link_status(self) -> Union[int, float]:
        """
        Returns the link status that is set when the condition is fullfilled.

        Returns
        -------
        `int` or `float`
            Pump speed if the link is a pump, otherwise one of the followig constants defined in
            :class:`~epyt_flow.simulation.events.actuator_events.ActuatorConstants`:

                - EN_CLOSED       = 0
                - EN_OPEN         = 1
                - EN_SET_CLOSED   = -1e10
                - EN_SET_OPEN     = 1e10

        """
        return self.__link_status

    @property
    def cond_type(self) -> int:
        """
        Returns the condition/rule type.

        Returns
        -------
        `int`
            Condition/Rule type -- will be one of the following EPANET constants:

                - EN_LOWLEVEL  = 0
                - EN_HILEVEL   = 1
                - EN_TIMER     = 2
                - EN_TIMEOFDAY = 3
        """
        return self.__cond_type

    @property
    def cond_var_value(self) -> Union[str, int]:
        """
        Return the condition/rule variable value.

        Node ID in the cases of EN_LOWLEVEL or EN_HILEVEL.
        Time of the day (in AM/PM format) in the case of EN_TIMEOFDAY.
        Number of hours (as an integer) since simulation start in the case of EN_TIMER.

        Returns
        -------
        `str` or `int`
            Condition/rule variable value.
        """
        return self.__cond_var_value

    @property
    def cond_comp_value(self) -> float:
        """
        Returns the condition/rule comparison value -- might be None if not needed.

        Lower or upper value on the pressure (or tank level) in the cases of
        EN_LOWLEVEL and EN_HILEVEL.

        Returns
        -------
        `float`
            Condition/Rule comparison value.
        """
        return self.__cond_comp_value

    def get_attributes(self) -> dict:
        return super().get_attributes() | {"link_id": self.__link_id,
                                           "link_status": self.__link_status,
                                           "cond_type": self.__cond_type,
                                           "cond_var_value": self.__cond_var_value,
                                           "cond_comp_value": self.__cond_comp_value}

    def __eq__(self, other) -> bool:
        return super().__eq__(other) and self.__link_id == other.link_id and \
            self.__link_status == other.link_status and self.__cond_type == other.cond_type and \
            self.__cond_var_value == other.cond_var_value and \
            self.__cond_comp_value == other.cond_comp_value

    def __str__(self) -> str:
        control_rule_str = f"LINK {self.__link_id} "
        if isinstance(self.__link_status, int):
            control_rule_str += "OPEN " if self.__link_status == ActuatorConstants.EN_OPEN or \
                self.__link_status == ActuatorConstants.EN_SET_OPEN else "CLOSED "
        else:
            control_rule_str += f"{self.__link_status} "

        if self.__cond_type == EpanetConstants.EN_TIMER:
            control_rule_str += f"AT TIME {self.__cond_var_value}"
        elif self.__cond_type == EpanetConstants.EN_TIMEOFDAY:
            control_rule_str += f"AT CLOCKTIME {self.__cond_var_value}"
        elif self.__cond_type == EpanetConstants.EN_LOWLEVEL:
            control_rule_str += f"IF NODE {self.__cond_var_value} BELOW {self.__cond_comp_value}"
        elif self.__cond_type == EpanetConstants.EN_HILEVEL:
            control_rule_str += f"IF NODE {self.__cond_var_value} ABOVE {self.__cond_comp_value}"

        return control_rule_str


class SimplePumpSpeedTimeControl(SimpleControlModule):
    """
    A class for representing a simple control rule for setting the pump speed at some point in time.

    Parameters
    ----------
    pump_id : `str`
        Pump ID.
    pump_speed : `float`
        Pump speed.
    time : `str` or `int`
        Time of the day (in AM/PM format) in the case or
        number of hours (as an integer) since simulation start.
    """
    def __init__(self, pump_id: str, pump_speed: float, time: Union[str, int]):
        super().__init__(link_id=pump_id, link_status=pump_speed,
                         cond_type=EpanetConstants.EN_TIMER if isinstance(time, int)
                         else EpanetConstants.EN_TIMEOFDAY,
                         cond_var_value=time, cond_comp_value=None)


class SimplePumpSpeedConditionControl(SimpleControlModule):
    """
    A class for representing a simple IF-THEN control rule for setting the pump speed.

    Parameters
    ----------
    Parameters
    ----------
    pump_id : `str`
        Pump ID.
    pump_speed : `float`
        Pump speed.
    node_id : `str`
        Node ID.
    comp_type : `int`
        Comparison type -- must be one of the following EPANET constants:

            - EN_LOWLEVEL  = 0
            - EN_HILEVEL   = 1
    comp_value : `float`:
        Lower or upper value on the pressure (or tank level) at which this
        control rule is triggered.
    """
    def __init__(self, pump_id: str, pump_speed: float, node_id: str, comp_type: int,
                 comp_value: float):
        super().__init__(link_id=pump_id, link_status=pump_speed, cond_type=comp_type,
                         cond_var_value=node_id, cond_comp_value=comp_value)


class SimpleValveTimeControl(SimpleControlModule):
    """
    A class for representing a simple control rule for setting the valve status
    at some point in time.

    Parameters
    ----------
    valve_id : `str`
        valve ID.
    valve_status : `int`
        Valve status -- must be one of the followig costants defined in
        :class:`~epyt_flow.simulation.events.actuator_events.ActuatorConstants`:

            - EN_CLOSED       = 0
            - EN_OPEN         = 1
            - EN_SET_CLOSED   = -1e10
            - EN_SET_OPEN     = 1e10
    time : `str` or `int`
        Time of the day (in AM/PM format) in the case or
        number of hours (as an integer) since simulation start.
    """
    def __init__(self, valve_id: str, valve_status: int, time: Union[str, int]):
        super().__init__(link_id=valve_id, link_status=valve_status,
                         cond_type=EpanetConstants.EN_TIMER if isinstance(time, int)
                         else EpanetConstants.EN_TIMEOFDAY,
                         cond_var_value=time, cond_comp_value=None)


class SimpleValveConditionControl(SimpleControlModule):
    """
    A class for representing a simple IF-THEN control rule for setting the valve status.

    Parameters
    ----------
    valve_id : `str`
        valve ID.
    valve_status : `int`
        Valve status -- must be one of the followig costants defined in
        :class:`~epyt_flow.simulation.events.actuator_events.ActuatorConstants`:

            - EN_CLOSED       = 0
            - EN_OPEN         = 1
            - EN_SET_CLOSED   = -1e10
            - EN_SET_OPEN     = 1e10
    node_id : `str`
        Node ID.
    comp_type : `int`
        Comparison type -- must be one of the following EPANET constants:

            - EN_LOWLEVEL  = 0
            - EN_HILEVEL   = 1
    comp_value : `float`:
        Lower or upper value on the pressure (or tank level) at which this
        control rule is triggered.
    """
    def __init__(self, valve_id: str, valve_status: int, node_id: str, comp_type: int,
                 comp_value: float):
        super().__init__(link_id=valve_id, link_status=valve_status, cond_type=comp_type,
                         cond_var_value=node_id, cond_comp_value=comp_value)



"""
Module provides a base class for control environments.
"""
import os
import uuid
from abc import abstractmethod, ABC
from typing import Union
import warnings
from epanet_plus import EpanetConstants

from ..simulation import ScenarioSimulator, ScenarioConfig, ScadaData
from ..utils import get_temp_folder


class ScenarioControlEnv(ABC):
    """
    Base class for a control environment challenge.

    Parameters
    ----------
    scenario_config : :class:`~epyt_flow.simulation.scenario_config.ScenarioConfig`
        Scenario configuration.
    autoreset : `bool`, optional
        If True, environment is automatically reset if terminated.

        The default is False.
    reapply_uncertainties_at_reset : `bool`, optional
        If True, the uncertainties are re-applied to the original properties at each reset.

        The default is False.

    Attributes
    ----------
    _scenario_sim : :class:`~epyt_flow.simulation.scenario_simulator.ScenarioSimulator`, protected
        Scenario simulator of the control scenario.
    _scenario_config : :class:`~epyt_flow.simulation.scenario_config.ScenarioConfig`
        Scenario configuration.
    _sim_generator : Generator[Union[:class:`~epyt_flow.simulation.scada.scada_data.ScadaData`, dict], bool, None], protected
        Generator for running the step-wise simulation.
    _hydraulic_scada_data : :class:`~epyt_flow.simulation.scada.scada_data.ScadaData`, protected
        SCADA data from the hydraulic simulation -- only used if EPANET-MSX is used in the control scenario.
    """
    def __init__(self, scenario_config: ScenarioConfig, autoreset: bool = False,
                 reapply_uncertainties_at_reset: bool = False,
                 **kwds):
        if not isinstance(scenario_config, ScenarioConfig):
            raise TypeError("'scenario_config' must be an instance of " +
                            "'epyt_flow.simulation.ScenarioConfig' " +
                            "but not of '{type(scenario_config)}'")
        if not isinstance(autoreset, bool):
            raise TypeError("'autoreset' must be an instance of 'bool' " +
                            f"but not of '{type(autoreset)}'")
        if not isinstance(reapply_uncertainties_at_reset, bool):
            raise TypeError("")

        self._scenario_config = scenario_config
        self._scenario_sim = None
        self._sim_generator = None
        self.__autoreset = autoreset
        self._hydraulic_scada_data = None
        self.__reapply_uncertainties_at_reset = reapply_uncertainties_at_reset

        super().__init__(**kwds)

    @property
    def autoreset(self) -> bool:
        """
        True, if environment automatically resets after it terminated.

        Returns
        -------
        `bool`
            True, if environment automatically resets after it terminated.
        """
        return self.__autoreset

    @property
    def reapply_uncertainties_at_reset(self) -> bool:
        """
        True, if the uncertainties are re-applied to the original properties at each reset.

        Returns
        -------
        `bool`
            True, if the uncertainties are re-applied to the original properties at each reset.
        """
        return self.__reapply_uncertainties_at_reset

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def close(self) -> None:
        """
        Frees all resources.
        """
        try:
            if self._sim_generator is not None:
                next(self._sim_generator)
                self._sim_generator.send(True)
                self._sim_generator = None
        except StopIteration:
            pass

        if self._scenario_sim is not None:
            self._scenario_sim.close()
            self._scenario_sim = None

    def contains_events(self) -> bool:
        """
        Check if the scenario contains any events.

        Returns
        -------
        `bool`
            True is the scenario contains any events, False otherwise.
        """
        return len(self._scenario_config.system_events) != 0 or \
            len(self._scenario_config.sensor_reading_events) != 0

    def reset(self) -> Union[tuple[ScadaData, bool], ScadaData]:
        """
        Resets the environment (i.e. simulation).

        Returns
        -------
        :class:`~epyt_flow.simulation.scada.scada_data.ScadaData`
            Current SCADA data (i.e. sensor readings).
        """
        self.close()

        self._scenario_sim = ScenarioSimulator(
            scenario_config=self._scenario_config)

        if self._scenario_sim.f_msx_in is not None:
            # Run hydraulic simulation first
            hyd_export = os.path.join(get_temp_folder(), f"epytflow_env_MSX_{uuid.uuid4()}.hyd")
            sim = self._scenario_sim.run_hydraulic_simulation
            self._hydraulic_scada_data = sim(hyd_export=hyd_export,
                                             reapply_uncertainties=self.__reapply_uncertainties_at_reset)

            # Run advanced quality analysis (EPANET-MSX) on top of the computed hydraulics
            gen = self._scenario_sim.run_advanced_quality_simulation_as_generator
            self._sim_generator = gen(hyd_export, support_abort=True)
        else:
            gen = self._scenario_sim.run_hydraulic_simulation_as_generator
            self._sim_generator = gen(support_abort=True,
                                      reapply_uncertainties=self.__reapply_uncertainties_at_reset)

        return self._next_sim_itr()

    def _next_sim_itr(self) -> Union[tuple[ScadaData, bool], ScadaData]:
        try:
            next(self._sim_generator)
            scada_data, terminated = self._sim_generator.send(False)

            if self._scenario_sim.f_msx_in is not None:
                cur_time = int(scada_data.sensor_readings_time[0])
                cur_hyd_scada_data = self._hydraulic_scada_data.\
                    extract_time_window(cur_time, cur_time)
                scada_data.join(cur_hyd_scada_data)

            if self.autoreset is True:
                return scada_data
            else:
                return scada_data, terminated
        except StopIteration:
            if self.__autoreset is True:
                return self.reset()
            else:
                return None, True

    def set_pump_status(self, pump_id: str, status: int) -> None:
        """
        Sets the status of a pump.

        Parameters
        ----------
        pump_id : `str`
            ID of the pump for which the status is set.
        status : `int`
            New status of the pump -- either active (i.e. open) or inactive (i.e. closed).

            Must be one of the following constants defined in
            :class:`~epyt_flow.simulation.events.actuator_events.ActuatorConstants`:

                - EN_CLOSED  = 0
                - EN_OPEN    = 1
        """
        if self._scenario_sim.f_msx_in is not None:
            raise RuntimeError("Can not execute actions affecting the hydraulics "+
                               "when running EPANET-MSX")

        pump_link_idx = self._scenario_sim.epanet_api.get_link_idx(pump_id)
        pattern_idx = self._scenario_sim.epanet_api.getlinkvalue(pump_link_idx,
                                                                 EpanetConstants.EN_LINKPATTERN)
        if pattern_idx != 0:
            warnings.warn(f"Can not set pump state of pump {pump_id} because a pump pattern exists")
        else:
            self._scenario_sim.epanet_api.setlinkvalue(pump_link_idx, EpanetConstants.EN_STATUS,
                                                       status)

    def set_pump_speed(self, pump_id: str, speed: float) -> None:
        """
        Sets the speed of a pump.

        Parameters
        ----------
        pump_id : `str`
            ID of the pump for which the pump speed is set.
        speed : `float`
            New pump speed.
        """
        if self._scenario_sim.f_msx_in is not None:
            raise RuntimeError("Can not execute actions affecting the hydraulics "+
                               "when running EPANET-MSX")

        pump_idx = self._scenario_sim.epanet_api.get_link_idx(pump_id)
        pattern_idx = int(self._scenario_sim.epanet_api.getlinkvalue(pump_idx,
                                                                     EpanetConstants.EN_LINKPATTERN))

        if pattern_idx == 0:
            warnings.warn(f"No pattern for pump '{pump_id}' found -- a new pattern is created")
            pattern_id = f"pump_speed_{pump_id}"
            self._scenario_sim.epanet_api.add_pattern(pattern_id, [speed])
            pattern_idx = self._scenario_sim.epanet_api.getpatternindex(pattern_id)
            self._scenario_sim.epanet_api.setlinkvalue(pump_idx, EpanetConstants.EN_LINKPATTERN,
                                                       pattern_idx)

        self._scenario_sim.epanet_api.setpattern(pattern_idx, [speed], 1)

    def set_valve_status(self, valve_id: str, status: int) -> None:
        """
        Sets the status of a valve.

        Parameters
        ----------
        valve_id : `str`
            ID of the valve for which the status is set.
        status : `int`
            New status of the valve -- either open or closed.

            Must be one of the following constants defined in
            :class:`~epyt_flow.simulation.events.actuator_events.ActuatorConstants`:

                - EN_CLOSED  = 0
                - EN_OPEN    = 1
        """
        if self._scenario_sim.f_msx_in is not None:
            raise RuntimeError("Can not execute actions affecting the hydraulics "+
                               "when running EPANET-MSX")

        valve_link_idx = self._scenario_sim.epanet_api.get_link_idx(valve_id)
        self._scenario_sim.epanet_api.setlinkvalue(valve_link_idx, EpanetConstants.EN_STATUS,
                                                   status)

    def set_node_quality_source_value(self, node_id: str, pattern_id: str,
                                      qual_value: float) -> None:
        """
        Sets the quality source at a particular node to a specific value -- e.g.
        setting the chlorine concentration injection to a specified value.

        Parameters
        ----------
        node_id : `str`
            ID of the node.
        pattern_id : `str`
            ID of the quality pattern at the specific node.
        qual_value : `float`
            New quality source value.
        """
        if self._scenario_sim.f_msx_in is not None:
            raise RuntimeError("Can not execute actions affecting the hydraulics "+
                               "when running EPANET-MSX")

        node_idx = self._scenario_sim.epanet_api.get_node_idx(node_id)
        pattern_idx = self._scenario_sim.epanet_api.getpatternindex(pattern_id)
        self._scenario_sim.epanet_api.setnodevalue(node_idx, EpanetConstants.EN_SOURCEQUAL, 1)
        self._scenario_sim.epanet_api.set_pattern(pattern_idx, [qual_value])

    def set_node_species_source_value(self, species_id: str, node_id: str, source_type: int,
                                      pattern_id: str, source_strength: float) -> None:
        """
        Sets the species source at a particular node to a specific value -- i.e.
        setting the species injection amount at a particular location.

        Parameters
        ----------
        species_id : `str`
            ID of the species.
        node_id : `str`
            ID of the node.
        source_type : `int`
            Type of the external species injection source -- must be one of
            the following EPANET constants:

                - EN_CONCEN     = 0
                - EN_MASS       = 1
                - EN_SETPOINT   = 2
                - EN_FLOWPACED  = 3

            Description:

                - E_CONCEN Sets the concentration of external inflow entering a node
                - EN_MASS Injects a given mass/minute into a node
                - EN_SETPOINT Sets the concentration leaving a node to a given value
                - EN_FLOWPACED Adds a given value to the concentration leaving a node
        pattern_id : `str`
            ID of the source pattern.
        source_strength : `float`
            Amount of the injected species (source strength) --
            i.e. interpreation of this number depends on `source_type`
        """
        if self._scenario_sim.f_msx_in is None:
            raise RuntimeError("You are not running EPANET-MSX")

        pattern_idx = self._scenario_sim.epanet_api.MSXgetindex(EpanetConstants.MSX_PATTERN,
                                                                pattern_id)
        self._scenario_sim.epanet_api.MSXsetpattern(pattern_idx, [1], 1)

        node_idx = self._scenario_sim.epanet_api.get_node_idx(node_id)
        species_idx = self._scenario_sim.epanet_api.get_msx_species_idx(species_id)
        self._scenario_sim.epanet_api.MSXsetsource(node_idx, species_idx, source_type,
                                                   source_strength, pattern_idx)

    @abstractmethod
    def step(self, *actions) -> Union[tuple[ScadaData, float, bool], tuple[ScadaData, float]]:
        """
        Performs the next step by applying an action and observing
        the consequences (SCADA data, reward, terminated).

        Note that `terminated` is only returned if `autoreset=False` otherwise
        only the current SCADA data and reward are returned.

        Returns
        -------
        `(` :class:`~epyt_flow.simulation.scada.scada_data.ScadaData` `, float, bool)` or `(` :class:`~epyt_flow.simulation.scada.scada_data.ScadaData` `, float)`
            Triple or tuple of observations (:class:`~epyt_flow.simulation.scada.scada_data.ScadaData`),
            reward (`float`), and terminated (`bool`).
        """
        raise NotImplementedError()



"""
Module provides functions and classes for serialization.
"""
from typing import Any, Union
from abc import abstractmethod, ABC
from io import BufferedIOBase
import pathlib
import importlib
import json
import gzip
import umsgpack
import numpy as np
import networkx
import scipy


SCIPY_BSRARRAY_ID                       = -3
NETWORKX_GRAPH_ID                       = -2
NUMPY_ARRAY_ID                          = -1
SENSOR_CONFIG_ID                        = 0
SCENARIO_CONFIG_ID                      = 1
MODEL_UNCERTAINTY_ID                    = 2
SENSOR_NOISE_ID                         = 3
ABSOLUTE_GAUSSIAN_UNCERTAINTY_ID        = 4
RELATIVE_GAUSSIAN_UNCERTAINTY_ID        = 5
ABSOLUTE_UNIFORM_UNCERTAINTY_ID         = 6
RELATIVE_UNIFORM_UNCERTAINTY_ID         = 7
PERCENTAGE_DEVIATON_UNCERTAINTY_ID      = 8
ABSOLUTE_DEEP_UNIFORM_UNCERTAINTY_ID    = 9
RELATIVE_DEEP_UNIFORM_UNCERTAINTY_ID    = 10
ABSOLUTE_DEEP_GAUSSIAN_UNCERTAINTY_ID   = 11
RELATIVE_DEEP_GAUSSIAN_UNCERTAINTY_ID   = 12
ABSOLUTE_DEEP_UNCERTAINTY_ID            = 13
RELATIVE_DEEP_UNCERTAINTY_ID            = 14
SENSOR_FAULT_CONSTANT_ID                = 15
SENSOR_FAULT_DRIFT_ID                   = 16
SENSOR_FAULT_GAUSSIAN_ID                = 17
SENSOR_FAULT_PERCENTAGE_ID              = 18
SENSOR_FAULT_STUCKATZERO_ID             = 19
LEAKAGE_ID                              = 20
ABRUPT_LEAKAGE_ID                       = 21
INCIPIENT_LEAKAGE_ID                    = 22
SCADA_DATA_ID                           = 23
SENSOR_ATTACK_OVERRIDE_ID               = 24
SENSOR_ATTACK_REPLAY_ID                 = 25
NETWORK_TOPOLOGY_ID                     = 26
PUMP_STATE_EVENT_ID                     = 28
PUMP_SPEED_EVENT_ID                     = 29
VALVE_STATE_EVENT_ID                    = 30
SPECIESINJECTION_EVENT_ID               = 31
SIMPLE_CONTROL_ID                       = 32
COMPLEX_CONTROL_ID                      = 33
COMPLEX_CONTROL_CONDITION_ID            = 34
COMPLEX_CONTROL_ACTION_ID               = 35
COLOR_SCHEMES_ID                        = 36


def my_packb(data: Any) -> bytes:
    """
    Overriden `umsgpack.packb <https://msgpack-python.readthedocs.io/en/latest/api.html#msgpack.packb>`_
    method to support custom serialization handlers.
    """
    return umsgpack.packb(data, ext_handlers=ext_handler_pack)


def my_unpackb(data: Any) -> Any:
    """
    Overriden `umsgpack.unpackb <https://msgpack-python.readthedocs.io/en/latest/api.html#msgpack.unpackb>`_
    method to support custom serialization handlers.
    """
    return umsgpack.unpackb(data, ext_handlers=ext_handler_unpack)


def serializable(my_id: int, my_file_ext: str) -> Any:
    """
    Decorator for a serializable class -- i.e. subclass of
    :class:`~epyt_flow.serialization.Serializable`.

    This decorator registers a new class as a serializable class.

    Parameters
    ----------
    my_id : `int`
        ID of the class.
    my_file_ext : `str`
        File extension.
    """
    def wrapper(my_class):
        @staticmethod
        def unpackb(data: bytes) -> Any:
            return my_class(**my_unpackb(data))
        setattr(my_class, "unpackb", unpackb)

        @staticmethod
        def file_ext() -> str:
            return my_file_ext
        setattr(my_class, "file_ext", file_ext)

        return umsgpack.ext_serializable(my_id)(my_class)

    return wrapper


class Serializable(ABC):
    """
    Base class for a serializable class -- must be used in conjunction with the
    :func:`~epyt_flow.serialization.serializable` decorator.
    """
    def __init__(self, _parent_path: str = "", **kwds):
        self._parent_path = _parent_path

        super().__init__(**kwds)

    @abstractmethod
    def get_attributes(self) -> dict:
        """
        Gets all attributes to be serialized -- these attributes are passed to the
        constructor when the object is deserialized.

        Returns
        -------
        `dict`
            Dictionary of attributes -- i.e. pairs of attribute name + value.
        """
        return {}

    def file_ext(self) -> str:
        """
        Returns the file extension of this class.

        This function is automatically implemented by applying the
        :func:`~epyt_flow.serialization.serializable` decorator.

        Returns
        -------
        `str`
            File extension.
        """
        raise NotImplementedError()

    def packb(self) -> bytes:
        """
        Serializes the attributes of this object.

        Returns
        -------
        `bytes`
            Serialized object.
        """
        return my_packb(self.get_attributes())

    @staticmethod
    def load(data: Union[bytes, BufferedIOBase]) -> Any:
        """
        Deserializes an instance of this class.

        Parameters
        ----------
        data : `bytes` or `io.BufferedIOBase`
            Serialized data or stream from which serialized data can be read.

        Returns
        -------
        `Any`
            Deserialized object.
        """
        return load(data)

    @staticmethod
    def load_from_file(f_in: str, use_zip: bool = True) -> Any:
        """
        Deserializes an instance of this class from a (compressed) file.

        Parameters
        ----------
        f_in : `str`
            Path to the file from which to deserialize the object.
        use_zip : `bool`, optional
            If True, the file `f_in` is supposed to be zip compressed -- False,
            if no compression was used when serializing the object.

            The default is True.

        Returns
        -------
        `Any`
            Deserialized object.
        """
        return load_from_file(f_in, use_zip)

    def dump(self, stream_out: BufferedIOBase = None) -> Any:
        """
        Serializes this object to a byte array.

        Parameters
        ----------
        stream_out : `io.BufferedIOBase`, optional
            Stream to which the serialized object is written.
            If None, the serialized object is returned as a `bytes` array.

            The default is None.

        Returns
        -------
        `bytes`
            If `stream_out` is None, the serialized object is returned as a `bytes` array.
        """
        return dump(self, stream_out)

    def save_to_file(self, f_out: str, use_zip: bool = True) -> None:
        """
        Serializes this instance and stores it in a (compressed) file.

        Parameters
        ----------
        f_in : `str`
            Path to the file where this serialized object will be stored.
        use_zip : `bool`, optional
            If True, the file `f_in` will be zip compressed -- False,
            if no compression is wanted.

            The default is True.
        """
        if not f_out.endswith(self.file_ext()):
            f_out += self.file_ext()

        return save_to_file(f_out, self, use_zip)


def my_to_json(obj: Any) -> str:
    """
    Serializes a given object to JSON.

    Parameters
    ----------
    obj : `Any`
        Object to be serialized.

    Returns
    -------
    `str`
        JSON data.
    """
    def __json_serialize(obj_: Any) -> dict:
        if isinstance(obj_, JsonSerializable):
            my_class_name = (obj_.__module__, obj_.__class__.__name__)
            return obj_.get_attributes() | {"__type__": my_class_name}
        elif isinstance(obj_, np.ndarray):
            return obj_.tolist()
        else:
            return obj_

    return json.dumps(obj, default=__json_serialize)


def my_load_from_json(data: str) -> Any:
    """
    Loads (i.e. deserializes) an object from given JSON data.

    Parameters
    ----------
    data : `str`
        JSON data.

    Returns
    -------
    `Any`
        Deserialized object.
    """
    def __object_hook(obj: dict) -> dict:
        if "__type__" in obj:
            module_name, class_name = obj["__type__"]
            cls = getattr(importlib.import_module(module_name), class_name)
            del obj["__type__"]

            for attr in obj:
                if isinstance(attr, dict):
                    obj[attr] = __object_hook(obj[attr])

            return cls(**obj)
        return obj

    return json.loads(data, object_hook=__object_hook)


class JsonSerializable(Serializable):
    """
    Base class for JSON serializable classes.
    Inherits from :class:`~epyt_flow.serialization.Serializable`.
    """

    def to_json(self) -> str:
        """
        Serializes this instance to JSON.

        Returns
        -------
        `str`
            JSON data.
        """
        return my_to_json(self)

    @staticmethod
    def load_from_json(data: str) -> Any:
        """
        Loads (i.e. deserializes) an instance of this class from given JSON data.

        Parameters
        ----------
        data : `str`
            JSON data.

        Returns
        -------
        `Any`
            Deserialized instance of this class.
        """
        return my_load_from_json(data)

    @staticmethod
    def load_from_json_file(f_in: str) -> Any:
        """
        Deserializes an instance of this class from a JSON file.

        Parameters
        ----------
        f_in : `str`
            Path to the JSON file from which to deserialize the object.

        Returns
        -------
        `Any`
            Deserialized object.
        """
        with open(f_in, "r", encoding="utf-8") as f:
            return my_load_from_json(f.read())

    def save_to_json_file(self, f_out: str) -> None:
        """
        Serializes this instance and stores it in a JSON file.

        Parameters
        ----------
        f_in : `str`
            Path to the JSON file where this serialized object will be stored.
        """
        if not f_out.endswith(self.file_ext()):
            f_out += self.file_ext()

        with open(f_out, "w", encoding="utf-8") as f:
            f.write(self.to_json())


def load(data: Union[bytes, BufferedIOBase]) -> Any:
    """
    Deserializes data.

    Parameters
    ----------
    data : `bytes` or `io.BufferedIOBase`
        Serialized data or stream from which serialized data can be read.

    Returns
    -------
    `Any`
        Deserialized data.
    """
    if isinstance(data, bytes):
        return my_unpackb(data)
    elif isinstance(data, BufferedIOBase):
        return my_unpackb(data.read())
    else:
        raise TypeError("Invalid type of 'data' -- must be either instance of 'bytes' or " +
                        f"'io.BufferedIOBase' but not of '{type(data)}'")


def dump(data: Any, stream_out: BufferedIOBase = None) -> Union[bytes, None]:
    """
    Serializes some given data to a byte array.

    Parameters
    ----------
    stream_out : `io.BufferedIOBase`, optional
        Stream to which the serialized object is written.
        If None, the serialized object is returned as a `bytes` array.

        The default is None.

    Returns
    -------
    `bytes`
        Serialized data if `stream_out` is None -- otherwise, nothing is returned.
    """
    if stream_out is None:
        return my_packb(data)
    else:
        if not isinstance(stream_out, BufferedIOBase):
            raise TypeError("'stream_out' must be an instance of 'io.BufferedIOBase' " +
                            f"but not of '{type(stream_out)}'")

        stream_out.write(my_packb(data))


def load_from_file(f_in: str, use_compression: bool = True) -> Any:
    """
    Deserializes data from a (compressed) file.

    Parameters
    ----------
    f_in : `str`
        Path to the file from which to deserialize the data.
    use_compression : `bool`, optional
        If True, the file `f_in` is supposed to be gzip compressed -- False,
        if no compression was used when serializing the data.

        The default is True.

    Returns
    -------
    `Any`
        Deserialized data.
    """
    inst = None

    if use_compression is False:
        with open(f_in, "rb") as f:
            inst = load(f.read())
    else:
        with gzip.open(f_in, "rb") as f:
            inst = load(f.read())

    if isinstance(inst, Serializable):
        inst._parent_path = pathlib.Path(f_in).parent.resolve()

    return inst


def save_to_file(f_out: str, data: Any, use_compression: bool = True) -> None:
    """
    Serializes data and stores it in a (compressed) file.

    Parameters
    ----------
    f_in : `str`
        Path to the file where the serialized data will be stored.
    use_compression : `bool`, optional
        If True, the file `f_in` will be gzip compressed -- False, if no compression is wanted.

        The default is True.
    """
    if use_compression is False:
        with open(f_out, "wb") as f:
            umsgpack.pack(data, f, ext_handlers=ext_handler_pack)
    else:
        with gzip.open(f_out, "wb") as f:
            f.write(dump(data))


# Add numpy.ndarray, networkx.Graph, and scipy.sparse.bsr_array support
def __encode_bsr_array(array: scipy.sparse.bsr_array
                       ) -> tuple[tuple[int, int], tuple[list[float], tuple[list[int], list[int]]]]:
    shape = array.shape
    data = []
    rows = []
    cols = []

    array_ = array.tocsr()   # Bug workaround: BSR arrays do not implement __getitem__
    for i, j in zip(*array_.nonzero()):
        rows.append(int(i))
        cols.append(int(j))
        data.append(float(array_[i, j]))

    return shape, (data, (rows, cols))


def __decode_bsr_array(ext_data: tuple[tuple[int, int],
                                       tuple[list[float], tuple[list[int], list[int]]]]
                       ) -> scipy.sparse.bsr_array:
    shape, data = ext_data
    return scipy.sparse.bsr_array((data[0], (data[1][0], data[1][1])), shape=(shape[0], shape[1]))


ext_handler_pack = {np.ndarray:
                    lambda arr: umsgpack.Ext(NUMPY_ARRAY_ID, umsgpack.packb(arr.tolist())),
                    networkx.Graph:
                        lambda graph:
                            umsgpack.Ext(NETWORKX_GRAPH_ID,
                                         umsgpack.packb(networkx.node_link_data(graph))),
                    scipy.sparse.bsr_array:
                    lambda arr: umsgpack.Ext(SCIPY_BSRARRAY_ID,
                                             umsgpack.packb(__encode_bsr_array(arr)))}
ext_handler_unpack = {NUMPY_ARRAY_ID: lambda ext: np.array(umsgpack.unpackb(ext.data)),
                      NETWORKX_GRAPH_ID:
                      lambda ext: networkx.node_link_graph(umsgpack.unpackb(ext.data)),
                      SCIPY_BSRARRAY_ID: lambda ext: __decode_bsr_array(umsgpack.unpackb(ext.data))}



"""
Module provides a class for representing the topology of WDN.
"""
from copy import deepcopy
import warnings
from typing import Any
import numpy as np
import networkx as nx
from scipy.sparse import bsr_array
from geopandas import GeoDataFrame
from shapely.geometry import Point, LineString
from epanet_plus import EpanetConstants

from .serialization import serializable, JsonSerializable, NETWORK_TOPOLOGY_ID


UNITS_USCUSTOM = 0
UNITS_SIMETRIC = 1


def unitscategoryid_to_str(unit_category_id: int) -> str:
    """
    Converts a given units category ID to the corresponding description.

    Parameters
    ----------
    unit_category_id : `int`
        ID of the units category.

        Must be one of the following constants:

            - UNITS_USCUSTOM = 0
            - UNITS_SIMETRIC = 1

    Returns
    -------
    `str`
        Units category description.
    """
    if unit_category_id is None:
        return ""
    elif unit_category_id == UNITS_USCUSTOM:
        return "US CUSTOMARY"
    elif unit_category_id == UNITS_SIMETRIC:
        return "SI METRIC"
    else:
        raise ValueError(f"Unknown units category ID '{unit_category_id}'")


@serializable(NETWORK_TOPOLOGY_ID, ".epytflow_topology")
class NetworkTopology(nx.Graph, JsonSerializable):
    """
    Class representing the topology of a WDN.

    Parameters
    ----------
    f_inp : `str`
        Path to .inp file to which this topology belongs.
    nodes : `list[tuple[str, dict]]`
        List of all nodes -- i.e. node ID and node information such as type and elevation.
    links : `list[tuple[str, tuple[str, str], dict]]`
        List of all links/pipes -- i.e. link ID, ID of connecting nodes, and link information
        such as pipe diameter, length, etc.
    pumps : `dict`
        List of all pumps -- i.e. valve ID, and information such as
        pump type and connecting nodes.
    valves : `dict`
        List of all valves -- i.e. valve ID, and information such as
        valve type and connecting nodes.
    units : `int`
        Measurement units category -- i.e. US Customary or SI Metric.

        Must be one of the following constants:

            - UNITS_USCUSTOM = 0  (US Customary)
            - UNITS_SIMETRIC = 1  (SI Metric)
    """
    def __init__(self, f_inp: str, nodes: list[tuple[str, dict]],
                 links: list[tuple[str, tuple[str, str], dict]],
                 pumps: dict,
                 valves: dict,
                 units: int,
                 **kwds):
        super().__init__(name=f_inp, **kwds)

        self.__nodes = nodes
        self.__links = links
        self.__pumps = pumps
        self.__valves = valves
        self.__units = units

        for node_id, node_info in nodes:
            node_elevation = node_info["elevation"]
            node_type = node_info["type"]
            self.add_node(node_id, info={"elevation": node_elevation, "type": node_type})

        for link_id, link, link_info in links:
            link_type = link_info["type"]
            link_diameter = link_info["diameter"]
            link_length = link_info["length"]
            self.add_edge(link[0], link[1], length=link_length,
                          info={"id": link_id, "type": link_type, "nodes": link,
                                "diameter": link_diameter, "length": link_length})

    def convert_units(self, units: int) -> Any:
        """
        Converts this instance to a :class:`~epyt_flow.topology.NetworkTopology` instance
        where everything is measured in given measurement units category
        (US Customary or SI Metric).

        Parameters
        ----------
        units : `int`
            Measurement units category.

            Must be one of the following constants:

                - UNITS_USCUSTOM = 0  (US Customary)
                - UNITS_SIMETRIC = 1  (SI Metric)

        Returns
        -------
        :class:`~epyt_flow.topology.NetworkTopology`
            Network topology with the new measurements units.
        """
        if self.__units is None:
            raise ValueError("This instance does not contain any units!")

        if not isinstance(units, int):
            raise TypeError(f"'units' must be an instance of 'int' but not of '{type(units)}'")
        if units not in [UNITS_SIMETRIC, UNITS_USCUSTOM]:
            raise ValueError(f"Invalid units '{units}'")

        if units == self.__units:
            warnings.warn("Units already set in this NetworkTopology instance -- nothing to do!")
            return deepcopy(self)

        # Get all data and convert units
        inch_to_millimeter = 25.4
        feet_to_meter = 0.3048

        nodes = []
        for node_id in self.get_all_nodes():
            node_info = self.get_node_info(node_id)
            if units == UNITS_USCUSTOM:
                conv_factor = 1. / feet_to_meter
            else:
                conv_factor = feet_to_meter
            node_info["elevation"] *= conv_factor
            if "diameter" in node_info:
                node_info["diameter"] *= conv_factor

            nodes.append((node_id, node_info))

        links = []
        for link_id, link_nodes in self.get_all_links():
            link_info = self.get_link_info(link_id)

            if units == UNITS_USCUSTOM:
                conv_factor = 1. / feet_to_meter
            else:
                conv_factor = feet_to_meter
            link_info["length"] *= conv_factor

            if units == UNITS_USCUSTOM:
                conv_factor = 1. / inch_to_millimeter
            else:
                conv_factor = inch_to_millimeter
            link_info["diameter"] *= conv_factor

            links.append((link_id, link_nodes, link_info))

        return NetworkTopology(f_inp=self.name, nodes=nodes, links=links, pumps=self.pumps,
                               valves=self.valves, units=units)

    def get_all_nodes(self) -> list[str]:
        """
        Gets a list of all nodes.

        Returns
        -------
        `list[str]`
            List of all nodes ID.
        """
        return [node_id for node_id, _ in self.__nodes]

    def get_number_of_nodes(self) -> int:
        """
        Returns the number of nodes.

        Returns
        -------
        `int`
            Number of nodes.
        """
        return len(self.get_all_nodes())

    def get_all_links(self) -> list[tuple[str, tuple[str, str]]]:
        """
        Gets a list of all links/pipes (incl. their end points).

        Returns
        -------
        `list[tuple[str, tuple[str, str]]]`
            List of links -- (link ID, (left node ID, right node ID)).
        """
        return [(link_id, end_points) for link_id, end_points, _ in self.__links]

    def get_number_of_links(self) -> int:
        """
        Returns the number of links.

        Returns
        -------
        `int`
            Number of links.
        """
        return len(self.get_all_links())

    def get_all_junctions(self) -> list[str]:
        """
        Gets all junctions -- i.e. nodes that are not tanks or reservoirs.

        Returns
        -------
        `list[str]`
            List of all junctions.
        """
        r = []

        for node_id in self.get_all_nodes():
            if self.get_node_info(node_id)["type"] == EpanetConstants.EN_JUNCTION:
                r.append(node_id)

        return r

    def get_number_of_junctions(self) -> int:
        """
        Returns the number of junctions.

        Returns
        -------
        `int`
            Number of junctions.
        """
        return len(self.get_all_junctions())

    def get_all_tanks(self) -> list[str]:
        """
        Gets all tanks -- i.e. nodes that are not junctions or reservoirs.

        Returns
        -------
        `list[str]`
            List of all tanks.
        """
        r = []

        for node_id in self.get_all_nodes():
            if self.get_node_info(node_id)["type"] == EpanetConstants.EN_TANK:
                r.append(node_id)

        return r

    def get_number_of_tanks(self) -> int:
        """
        Returns the number of tanks.

        Returns
        -------
        `int`
            Number of tanks.
        """
        return len(self.get_all_tanks())

    def get_all_reservoirs(self) -> list[str]:
        """
        Gets all reservoirs -- i.e. nodes that are not junctions or tanks.

        Returns
        -------
        `list[str]`
            List of all reservoirs.
        """
        r = []

        for node_id in self.get_all_nodes():
            if self.get_node_info(node_id)["type"] == EpanetConstants.EN_RESERVOIR:
                r.append(node_id)

        return r

    def get_number_of_reservoirs(self) -> int:
        """
        Returns the number of reservoirs.

        Returns
        -------
        `int`
            Number of reservoirs.
        """
        return len(self.get_all_reservoirs())

    def get_all_pipes(self) -> list[tuple[str, tuple[str, str]]]:
        """
        Gets all pipes -- i.e. links that not valves or pumps.

        Returns
        -------
        `list[tuple[str, tuple[str, str]]]`
            List of all pipes -- (link ID, (left node ID, right node ID)).
        """
        r = []

        for link_id, link_nodes in self.get_all_links():
            link_info = self.get_link_info(link_id)

            if link_info["type"] == EpanetConstants.EN_PIPE:
                r.append((link_id, link_nodes))

        return r

    def get_number_of_pipes(self) -> int:
        """
        Returns the number of pipes.

        Returns
        -------
        `int`
            Number of pipes.
        """
        return len(self.get_all_pipes())

    def get_all_pumps(self) -> list[str]:
        """
        Gets the IDs of all pumps.

        Returns
        -------
        `list[str]`
            Pump IDs.
        """
        return list(self.__pumps.keys())

    def get_number_of_pumps(self) -> int:
        """
        Returns the number of pumps.

        Returns
        -------
        `int`
            Number of pumps.
        """
        return len(self.get_all_pumps())

    def get_all_valves(self) -> list[str]:
        """
        Gets the IDs of all valves.

        Returns
        -------
        `list[str]`
            Valve IDs.
        """
        return list(self.__valves.keys())

    def get_number_of_valves(self) -> int:
        """
        Returns the number of valves.

        Returns
        -------
        `int`
            Number of valves.
        """
        return len(self.get_all_valves())

    def get_node_info(self, node_id: str) -> dict:
        """
        Gets all information (e.g. elevation, type, etc.) associated with a given node.

        Parameters
        ----------
        node_id : `str`
            ID of the node.

        Returns
        -------
        `dict`
            Information associated with the given node.
        """
        for node_id_, node_info in self.__nodes:
            if node_id_ == node_id:
                return node_info

        raise ValueError(f"Unknown node '{node_id}'")

    def get_link_info(self, link_id: str) -> dict:
        """
        Gets all information (e.g. diameter, length, etc.) associated with a given link.

        Note that links can be pipes, pumps, or valves.

        Parameters
        ----------
        link_id : `str`
            ID of the link.

        Returns
        -------
        `dict`
            Information associated with the given link.
        """
        for link_id_, link_nodes, link_info in self.__links:
            if link_id_ == link_id:
                return {"nodes": link_nodes} | link_info

        raise ValueError(f"Unknown link '{link_id}'")

    def get_pump_info(self, pump_id: str) -> dict:
        """
        Gets all information associated with a given pump.

        Parameters
        ----------
        pump_id : `str`
            ID of the pump.

        Returns
        -------
        `dict`
            Pump information.
        """
        if pump_id in self.__pumps:
            return self.__pumps[pump_id]
        else:
            raise ValueError(f"Unknown pump: '{pump_id}'")

    def get_valve_info(self, valve_id: str) -> dict:
        """
        Gets all information associated with a given valve.

        Parameters
        ----------
        valve_id : `str`
            ID of the valve.

        Returns
        -------
        `dict`
            Valve information.
        """
        if valve_id in self.__valves:
            return self.__valves[valve_id]
        else:
            raise ValueError(f"Unknown valve: '{valve_id}'")

    @property
    def pumps(self) -> dict:
        """
        Gets all pumps -- i.e. ID and associated information such as the pump type.

        Returns
        -------
        `dict`
            All pumps and their associated information.
        """
        return deepcopy(self.__pumps)

    @property
    def valves(self) -> dict:
        """
        Gets all valves -- i.e. ID and associated information such as the valve type.

        Returns
        -------
        `dict`
            All valves and their associated information.
        """
        return deepcopy(self.__valves)

    @property
    def units(self) -> int:
        """
        Gets the used measurement units category.

        Will be one of the following constants:

            - UNITS_USCUSTOM = 0  (US Customary)
            - UNITS_SIMETRIC = 1  (SI Metric)

        Returns
        -------
        `int`
            Measurement units category.
        """
        return self.__units

    def __eq__(self, other) -> bool:
        if not isinstance(other, NetworkTopology):
            raise TypeError("Can not compare 'NetworkTopology' instance to " +
                            f"'{type(other)}' instance")

        adj_matrix = self.get_adj_matrix()
        other_adj_matrix = other.get_adj_matrix()

        return super().__eq__(other) and \
            self.name == other.name \
            and not np.any(adj_matrix.data != other_adj_matrix.data) \
            and not np.any(adj_matrix.indices != other_adj_matrix.indices) \
            and not np.any(adj_matrix.indptr != other_adj_matrix.indptr) \
            and self.get_all_nodes() == other.get_all_nodes() \
            and all(link_a[0] == link_b[0] and link_a[1] == link_b[1]
                    for link_a, link_b in zip(self.get_all_links(), other.get_all_links())) \
            and self.__units == other.units \
            and self.get_all_pumps() == other.get_all_pumps() \
            and self.get_all_valves() == other.get_all_valves()

    def __str__(self) -> str:
        return f"f_inp: {self.name} nodes: {self.__nodes} links: {self.__links} " +\
            f"pumps: {self.__pumps} valves: {self.__valves} " +\
            f"units: {unitscategoryid_to_str(self.__units)}"

    def get_attributes(self) -> dict:
        return super().get_attributes() | {"f_inp": self.name,
                                           "nodes": self.__nodes,
                                           "links": self.__links,
                                           "pumps": self.__pumps,
                                           "valves": self.__valves,
                                           "units": self.__units}

    def to_gis(self, coord_reference_system: str = None, pumps_as_points: bool = False,
               valves_as_points: bool = False) -> dict:
        """
        Gets the network topology as a dictionary of `geopandas.GeoDataFrames` instances --
        i.e. each quantity (nodes, links/pipes, valves, etc.) is represented by a
        `geopandas.GeoDataFrames` instance.

        Parameters
        ----------
        coord_reference_system : `str`, optional
            Coordinate reference system.

            The default is None.
        pumps_as_points : `bool`, optional
            If True, pumps are represented by points, otherwise by lines.

            The default is False.

        valves_as_points : `bool`, optional
            If True, valves are represented by points, otherwise by lines.

            The default is False.

        Returns
        -------
        `dict`
            Network topology as a dictionary of `geopandas.GeoDataFrames <https://geopandas.org/en/stable/docs/reference/api/geopandas.GeoDataFrame.html>`_ instances.
            If a quantity does not exist, the data frame will be None.
        """
        gis = {"nodes": None, "links": None,
               "tanks": None, "reservoirs": None,
               "valves": None, "pumps": None}

        # Nodes
        node_data = {"id": [], "type": [], "elevation": [], "geometry": []}
        tank_data = {"id": [], "volume": [], "max_level": [], "min_level": [], "mixing_fraction": [],
                     "elevation": [], "diameter": [], "geometry": []}
        reservoir_data = {"id": [], "elevation": [], "geometry": []}
        for node_id in self.get_all_nodes():
            node_info = self.get_node_info(node_id)

            node_data["id"].append(node_id)
            node_data["type"].append(node_info["type"])
            node_data["elevation"].append(node_info["elevation"])
            node_data["geometry"].append(Point(node_info["coord"]))

            if node_info["type"] == EpanetConstants.EN_TANK:
                tank_data["id"].append(node_id)
                tank_data["elevation"].append(node_info["elevation"])
                tank_data["diameter"].append(node_info["diameter"])
                tank_data["volume"].append(node_info["volume"])
                tank_data["max_level"].append(node_info["max_level"])
                tank_data["min_level"].append(node_info["min_level"])
                tank_data["mixing_fraction"].append(node_info["mixing_fraction"])
                tank_data["mixing_model"].append(node_info["mixing_model"])
                tank_data["geometry"].append(Point(node_info["coord"]))
            elif node_info["type"] == EpanetConstants.EN_RESERVOIR:
                reservoir_data["id"].append(node_id)
                reservoir_data["elevation"].append(node_info["elevation"])
                reservoir_data["geometry"].append(Point(node_info["coord"]))

        gis["nodes"] = GeoDataFrame(node_data, crs=coord_reference_system)
        gis["tanks"] = GeoDataFrame(tank_data, crs=coord_reference_system)
        gis["reservoirs"] = GeoDataFrame(reservoir_data, crs=coord_reference_system)

        # Links
        pipe_data = {"id": [], "type": [], "end_point_a": [], "end_point_b": [],
                     "length": [], "diameter": [], "geometry": []}
        valve_data = {"id": [], "type": [], "geometry": []}
        pump_data = {"id": [], "type": [], "geometry": []}
        for link_id, link_nodes in self.get_all_links():
            link_info = self.get_link_info(link_id)
            end_points_coord = [self.get_node_info(n)["coord"] for n in link_nodes]

            if link_info["type"] == EpanetConstants.EN_PIPE:
                pipe_data["id"].append(link_id)
                pipe_data["type"].append(link_info["type"])
                pipe_data["end_point_a"].append(link_nodes[0])
                pipe_data["end_point_b"].append(link_nodes[1])
                pipe_data["length"].append(link_info["length"])
                pipe_data["diameter"].append(link_info["diameter"])
                pipe_data["geometry"].append(LineString(end_points_coord))
            elif link_info["type"] == EpanetConstants.EN_PUMP:
                pump_data["id"].append(link_id)
                pump_data["type"].append(self.get_pump_info(link_id)["type"])
                if pumps_as_points is True:
                    pump_data["geometry"].append(Point(end_points_coord[0]))
                else:
                    pump_data["geometry"].append(LineString(end_points_coord))
            else:   # Valve
                valve_data["id"].append(link_id)
                valve_data["type"].append(self.get_valve_info(link_id)["type"])
                if valves_as_points is True:
                    valve_data["geometry"].append(Point(end_points_coord[0]))
                else:
                    valve_data["geometry"].append(LineString(end_points_coord))

        gis["pipes"] = GeoDataFrame(pipe_data, crs=coord_reference_system)
        gis["valves"] = GeoDataFrame(valve_data, crs=coord_reference_system)
        gis["pumps"] = GeoDataFrame(pump_data, crs=coord_reference_system)

        return gis

    def get_adj_matrix(self) -> bsr_array:
        """
        Gets the adjacency matrix of this graph.

        Returns
        -------
        `scipy.bsr_array <https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.bsr_array.html>`_
            Adjacency matrix as a sparse array.
        """
        nodes_id = [node_id for node_id, _ in self.__nodes]
        n_nodes = len(self.__nodes)

        row = []
        col = []
        for _, link_end_points, _ in self.__links:
            a = nodes_id.index(link_end_points[0])
            b = nodes_id.index(link_end_points[1])

            row.append(a)
            col.append(b)

            row.append(b)
            col.append(a)

        for i in range(n_nodes):
            row.append(i)
            col.append(i)

        return bsr_array((np.ones(len(row)), (row, col)), shape=(n_nodes, n_nodes))

    def get_adj_list(self) -> dict[str, list[str]]:
        """
        Returns the connectivity of the nodes (node IDs) as an adjacency list.

        Returns
        -------
        `dict[str, list[str]]`
            Adjacency list as a dictionary.
        """
        adj_list = {}

        for node_id in self.get_all_nodes():
            adj_list[node_id] = self.get_neighbors(node_id)

        return adj_list

    def get_neighbors(self, node_id: str) -> list[str]:
        """
        Gets all neighboring nodes of a given node.

        Parameters
        ----------
        node_id : `str`
            ID of the node.

        Returns
        -------
        `list[str]`
            IDs of neighboring nodes.
        """
        if node_id not in self.get_all_nodes():
            raise ValueError(f"Unknown node '{node_id}'")

        return list(self.neighbors(node_id))

    def get_adjacent_links(self, node_id: str) -> list[tuple[str, tuple[str, str]]]:
        """
        Gets all adjacent links/pipes of a given node.

        Parameters
        ----------
        node_id : `str`
            ID of the node.

        Returns
        -------
        `list[tuple[str, tuple[str, str]]]`
            Adjacent links -- i.e. (link ID, IDs of node end points).
        """
        if node_id not in self.get_all_nodes():
            raise ValueError(f"Unknown node '{node_id}'")

        links = []

        for link_id, nodes_id, _ in self.__links:
            if node_id in nodes_id:
                links.append((link_id, nodes_id))

        return links

    def get_shortest_path(self, start_node_id: str, end_node_id: str,
                          use_pipe_length_as_weight: bool = True) -> list[str]:
        """
        Computes the shortest path between two nodes in this graph.

        Parameters
        ----------
        start_node_id : `str`
            ID of start node.
        end_node_id : `str`
            ID of end node.
        use_pipe_length_as_weight : `bool`, optional
            If True, pipe lengths are used for the edge weights -- otherwise,
            each edge weight is set to one.

            The default is True.
        """
        if start_node_id not in self.get_all_nodes():
            raise ValueError(f"Unknown node '{start_node_id}'")
        if end_node_id not in self.get_all_nodes():
            raise ValueError(f"Unknown node '{end_node_id}'")

        weight = "length" if use_pipe_length_as_weight is True else None
        return nx.shortest_path(self, source=start_node_id, target=end_node_id, weight=weight)

    def get_all_pairs_shortest_path(self, use_pipe_length_as_weight: bool = True) -> dict:
        """
        Computes the shortest path between all pairs of nodes in this graph.

        Parameters
        ----------
        use_pipe_length_as_weight : `bool`, optional
            If True, pipe lengths are used for the edge weights -- otherwise,
            each edge weight is set to one.

            The default is True.

        Returns
        -------
        `dict`
            Shortest paths between all pairs of nodes as nested dictionaries --
            first key is the start node, second key is the end node.
        """
        weight = "length" if use_pipe_length_as_weight is True else None
        return nx.shortest_path(self, weight=weight)

    def get_shortest_path_length(self, start_node_id: str, end_node_id: str,
                                 use_pipe_length_as_weight: bool = True) -> list[str]:
        """
        Computes the shortest path length between two nodes in this graph.

        Parameters
        ----------
        start_node_id : `str`
            ID of start node.
        end_node_id : `str`
            ID of end node.
        use_pipe_length_as_weight : `bool`, optional
            If True, pipe lengths are used for the edge weights -- otherwise,
            each edge weight is set to one.

            The default is True.
        """
        if start_node_id not in self.get_all_nodes():
            raise ValueError(f"Unknown node '{start_node_id}'")
        if end_node_id not in self.get_all_nodes():
            raise ValueError(f"Unknown node '{end_node_id}'")

        weight = "length" if use_pipe_length_as_weight is True else None
        return nx.shortest_path_length(self, source=start_node_id, target=end_node_id,
                                       weight=weight)

    def get_all_pairs_shortest_path_length(self, use_pipe_length_as_weight: bool = True) -> dict:
        """
        Computes the shortest path length between all pairs of nodes in this graph.

        Parameters
        ----------
        use_pipe_length_as_weight : `bool`, optional
            If True, pipe lengths are used for the edge weights -- otherwise,
            each edge weight is set to one.

            The default is True.

        Returns
        -------
        `dict`
            Shortest paths between all pairs of nodes as nested dictionaries --
            first key is the start node, second key is the end node.
        """
        weight = "length" if use_pipe_length_as_weight is True else None
        return dict(nx.shortest_path_length(self, weight=weight))
